import logging
import hashlib
import secrets
import string
from typing import Awaitable, Callable

from .type_int import TypeInt
from .message import MessageSend

logger = logging.getLogger(__name__)


def generate_key(length: int = 12) -> str:
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(length))


def hash_device(device_id: str, password: str, salt: str | None = None) -> dict:
    inner = hashlib.sha256((device_id + password).encode("utf-8")).hexdigest()

    if not isinstance(salt, str) or len(salt) == 0:
        salt = generate_key(12)

    final_hash = hashlib.sha256((salt + inner).encode("utf-8")).hexdigest()

    return {
        "salt": salt,
        "hash": final_hash,
    }


class Authenticate:
    def __init__(self, connection) -> None:
        self.connection = connection
        self.auto_start_session = True
        self.async_auth_result_callback: Callable[[bool], Awaitable[None]] | None = None

    def create_auth_msg(self, username: str, password: str) -> MessageSend:
        auth_data = hash_device(self.connection.device_id, password)
        return MessageSend(
            type_int=TypeInt.AUTH_LOGIN,
            mc=1,
            payload={
                "username": username,
                "password": auth_data["hash"],
                "salt": auth_data["salt"],
            },
        )

    async def enqueue_authentication_flow(
        self,
        username: str,
        password: str,
        auto_start_session: bool = True,
        async_auth_result_callback: Callable[[bool], Awaitable[None]] | None = None,
    ):
        self.auto_start_session = auto_start_session
        self.async_auth_result_callback = async_auth_result_callback
        auth_msg = self.create_auth_msg(username, password)
        await self.connection.enqueue_message(auth_msg)
        logger.debug("Queued AUTH_LOGIN [30]")

    async def execute_authentication_now(self, username: str, password: str) -> bool:
        auth_msg = self.create_auth_msg(username, password)
        await self.connection.send_message(auth_msg)
        logger.debug("Sent AUTH_LOGIN [30] immediately")

        response = await self.connection.receive_message()

        if response.type_int == TypeInt.AUTH_LOGIN_SUCCESS:
            payload = response.payload
            self.connection.token = payload.get("token")
            logger.debug("Authentication successful! Token received.")
            return True
        if response.type_int == TypeInt.AUTH_LOGIN_DENIED:
            logger.debug("Authentication denied - wrong username or password.")
            return False
        raise ValueError(
            f"Unexpected message type during authentication: {response.type_int}"
        )

    async def _async_auth_result_callback(self, success: bool):
        if self.async_auth_result_callback is None:
            return
        try:
            await self.async_auth_result_callback(success)
        except Exception:
            logger.exception("async_auth_result_callback failed")

    async def recv_AUTH_LOGIN_SUCCESS(self, message):
        payload = message.payload
        self.connection.token = payload.get("token")

        logger.debug("Login success! Token received.")

        await self._async_auth_result_callback(True)

        if self.auto_start_session:
            await self.enqueue_AUTH_APPLY_TOKEN()

    async def recv_AUTH_LOGIN_DENIED(self, message):
        logger.error("Login denied by the server (device) - wrong username or password.")
        await self._async_auth_result_callback(False)

    async def enqueue_AUTH_APPLY_TOKEN(self):
        await self.connection.enqueue_message(
            MessageSend(
                type_int=TypeInt.AUTH_APPLY_TOKEN,
                mc=-1,
                payload={"token": self.connection.token},
            )
        )
        logger.debug("Queued AUTH_APPLY_TOKEN [33]")

    async def recv_AUTH_APPLY_TOKEN_RESPONSE(self, message):
        payload = message.payload
        valid = payload.get("valid")
        remaining = payload.get("remaining")

        logger.debug(f"Auth Result: valid={valid}, remaining={remaining}")

        if not valid:
            logger.warning("Applying the token failed!")



"""Async client for Harreither Brain devices."""

from .authenticate import Authenticate, generate_key, hash_device
from .connection import Connection
from .receive import ReceiveData
from .establish_connection import EstablishConnection
from .message import MessageReceived, MessageSend
from .type_int import TypeInt

__all__ = [
    "Authenticate",
    "Connection",
    "ReceiveData",
    "EstablishConnection",
    "MessageReceived",
    "MessageSend",
    "TypeInt",
    "generate_key",
    "hash_device",
]

__version__ = "0.1.0"



# Special value indicating that the message counter should be auto-filled
MC_AUTO = -1000


class MessageReceived:
    def __init__(self, type_int, mc, payload, ref=None) -> None:
        self.type_int = type_int
        self.mc = mc
        self.payload = payload
        self.ref = ref

    def __repr__(self):
        return (
            "MessageReceived(type_int="
            f"{self.type_int}, mc={self.mc}, payload={self.payload}, ref={self.ref})"
        )


class MessageSend:
    def __init__(self, type_int, mc=None, payload=None, ref=None) -> None:
        self.type_int = type_int
        self.mc = mc
        self.payload = payload
        self.ref = ref

    def __repr__(self):
        return (
            "MessageSend(type_int="
            f"{self.type_int}, mc={self.mc}, payload={self.payload}, ref={self.ref})"
        )



import unittest

from harreither_brain_client.data import Data
from harreither_brain_client.message import MessageReceived
from harreither_brain_client.type_int import TypeInt


class DummyConnection:
    def __init__(self, strict: bool = False) -> None:
        self.strict = strict
        self.sent_acks = []
        self.notifications = []

    async def send_ack_message(self, message) -> None:
        self.sent_acks.append(message)

    async def async_notify_update(self, key, value_dict, new):
        self.notifications.append((key, value_dict, new))


class DataHandlerTests(unittest.IsolatedAsyncioTestCase):
    async def test_recv_add_screen_stores_screen_and_sends_ack(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        message = MessageReceived(
            type_int=TypeInt.ADD_SCREEN,
            mc=123,
            payload={
                "screen": {
                    "screenID": 42,
                    "title": "Diagnostics",
                    "statuspage": True,
                    "itemCount": 2,
                }
            },
        )

        await data.recv_ADD_SCREEN(message)

        self.assertIn(42, data.screens)
        self.assertEqual(data.screens[42]["title"], "Diagnostics")
        self.assertEqual(data.screens[42]["itemCount"], 2)
        self.assertEqual(connection.sent_acks, [message])

    async def test_recv_app_info_sends_ack(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        message = MessageReceived(
            type_int=TypeInt.APP_INFO,
            mc=5,
            payload={"info": "500"},
        )

        await data.recv_APP_INFO(message)

        self.assertEqual(connection.sent_acks, [message])

    async def test_recv_add_dbentries_stores_entries(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        message = MessageReceived(
            type_int=TypeInt.ADD_DBENTRIES,
            mc=6,
            payload={
                "DBentries": [
                    {"VID": 10, "type": 15, "text": "Arhiviranje"},
                    {"VID": 127, "type": 10, "text": "Cas"},
                ]
            },
        )

        await data.recv_ADD_DBENTRIES(message)

        self.assertIn(10, data.dbentries)
        self.assertEqual(data.dbentries[10]["text"], "Arhiviranje")
        self.assertIn(127, data.dbentries)
        self.assertEqual(connection.sent_acks, [message])

    async def test_recv_set_home_data_updates_device_info(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        message = MessageReceived(
            type_int=TypeInt.SET_HOME_DATA,
            mc=4,
            payload={
                "id": "home-1",
                "name": "Primary",
                "isvalid": True,
                "info": "506",
            },
        )

        await data.recv_SET_HOME_DATA(message)

        self.assertEqual(connection.device_home_id, "home-1")
        self.assertEqual(connection.device_home_name, "Primary")
        self.assertEqual(connection.sent_acks, [message])

    async def test_recv_add_items_creates_entries_and_sends_ack(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        data.dbentries[500] = {"VID": 500, "type": 10, "text": "Temp"}
        message = MessageReceived(
            type_int=TypeInt.ADD_ITEMS,
            mc=7,
            payload={
                "screenID": 100,
                "pos": 0,
                "items": [
                    {
                        "VID": 500,
                        "detail": 1,
                        "name": "Temperature",
                        "edit": False,
                        "value": 21.5,
                        "validity": 0,
                        "valstr": "21.5",
                    }
                ],
                "end": False,
            },
        )

        await data.recv_ADD_ITEMS(message)

        key = (500, 1, None)
        self.assertIn(key, data.entries._entries)
        self.assertEqual(
            data.entries._entries[key]["vid_obj"], {"VID": 500, "type": 10, "text": "Temp"}
        )
        self.assertEqual(data.screens[100]["items"][0]["name"], "Temperature")
        self.assertEqual(connection.sent_acks, [message])
        self.assertEqual(connection.notifications[0][0], key)
        self.assertTrue(connection.notifications[0][2])

    async def test_recv_set_alerts_restarts_and_appends(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        data.alerts = [{"text": "old", "type": 1, "icon": 1}]
        message = MessageReceived(
            type_int=TypeInt.SET_ALERTS,
            mc=8,
            payload={
                "restart": True,
                "alerts": [{"text": "1007", "type": 1, "icon": 1}],
                "end": True,
            },
        )

        await data.recv_SET_ALERTS(message)

        self.assertEqual(data.alerts, [{"text": "1007", "type": 1, "icon": 1}])
        self.assertEqual(connection.sent_acks, [message])

    async def test_recv_update_items_updates_existing_entries(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        key = (116, 30121, None)
        data.entries._entries[key] = {
            "VID": 116,
            "detail": 30121,
            "value": 40.0,
            "validity": 0,
        }
        message = MessageReceived(
            type_int=TypeInt.UPDATE_ITEMS,
            mc=9,
            payload={
                "items": [
                    {
                        "VID": 116,
                        "detail": 30121,
                        "value": 41.4,
                        "validity": 0,
                    }
                ],
                "end": True,
            },
        )

        await data.recv_UPDATE_ITEMS(message)

        self.assertEqual(data.entries._entries[key]["value"], 41.4)
        self.assertEqual(connection.sent_acks, [message])
        self.assertEqual(connection.notifications[0][0], key)
        self.assertEqual(connection.notifications[0][1]["value"], 41.4)
        self.assertFalse(connection.notifications[0][2])

    async def test_add_entries_handles_citems_recursively(self):
        connection = DummyConnection(strict=True)
        data = Data(connection)
        data.dbentries.update(
            {
                50005: {"VID": 50005, "text": "Group"},
                3: {"VID": 3, "text": "System"},
                122: {"VID": 122, "text": "Mode"},
            }
        )

        items = [
            {
                "VID": 50005,
                "detail": 2,
                "citems": [
                    {"VID": 3, "detail": 1001, "objID": 1, "name": "CELOTNI SISTEM"},
                    {
                        "VID": 122,
                        "detail": 1,
                        "edit": False,
                        "value": 1,
                        "validity": 0,
                        "valstr": "Zima",
                    },
                ],
            }
        ]

        await data.add_entries(items)

        parent_key = (50005, 2, 1)
        child_key = (3, 1001, 1)
        sibling_key = (122, 1, None)

        self.assertIn(parent_key, data.entries._entries)
        self.assertIn(child_key, data.entries._entries)
        self.assertIn(sibling_key, data.entries._entries)
        self.assertEqual(
            data.entries._entries[parent_key]["vid_obj"], {"VID": 50005, "text": "Group"}
        )
        self.assertEqual(
            data.entries._entries[child_key]["vid_obj"], {"VID": 3, "text": "System"}
        )
        self.assertEqual(
            data.entries._entries[sibling_key]["vid_obj"], {"VID": 122, "text": "Mode"}
        )
        self.assertEqual(len(connection.notifications), 3)
        self.assertTrue(all(n[2] for n in connection.notifications))


if __name__ == "__main__":
    unittest.main()



from enum import IntEnum


class TypeInt(IntEnum):
    NACK_INFO_UNKNOWN_DEVICE = -100
    NACK_INFO_DEVICE_NOT_DIMMABLE = -99
    NACK_INFO_INVALID_ACTION = -98
    IDLE = -1
    NACK = 0
    ACK = 1
    HEARTBEAT = 2
    WAIT4ACK = 3
    CONNECTION_START = 10
    CONNECTION_CONFIRM = 11
    CONNECTION_ESTABLISHED = 12
    CONNECTION_DECLINED = 13
    SC_INIT = 14
    SC_PUBKEY = 15
    SC_SECRET = 16
    SC_ESTABLISHED = 17
    SC_INVALID = 18
    AUTH_LOGIN = 30
    AUTH_LOGIN_DENIED = 31
    AUTH_LOGIN_SUCCESS = 32
    AUTH_APPLY_TOKEN = 33
    AUTH_APPLY_TOKEN_RESPONSE = 34
    AUTH_VERIFY_TOKEN = 35
    AUTH_VERIFY_TOKEN_RESPONSE = 36
    AUTH_RENEW_TOKEN = 37
    AUTH_RENEW_TOKEN_RESPONSE = 38
    AUTH_KILL_TOKEN = 39
    TEST_ON = 100
    TEST_OFF = 101
    TEST_CRYPTO = 102
    TEST_ALERT = 104
    TEST_DEVICE_STATE = 105
    TEST_SEND_FROM_APP = 107
    INITIAL_DATA = -100
    HOME_DATA = -101
    ACTUAL_SCREEN = 200
    ACTION_SELECTED = 201
    ACTION_EDITED_VALUE = 202
    ACTION_REQUEST_UI = 203
    INIT_SWUPDATE = 247
    DATA_SWUPDATE = 248
    START_SWUPDATE = 249
    SET_TIME = 250

    TEST_SEND_FROM_BRAIN = 106
    APP_INFO = 295
    ADD_SCREEN = 296
    ADD_DBENTRIES = 297
    ADD_ITEMS = 299
    UPDATE_ITEMS = 300
    SET_HOME_DATA = 301
    SET_ALERTS = 302

    @classmethod
    def get_name(cls, id_val):
        """Translate an integer ID to its symbolic name."""
        try:
            return cls(id_val).name
        except ValueError:
            return f"UNKNOWN_TYPE_{id_val}"



import asyncio
import base64
import json
import logging
from contextlib import suppress

import websockets

from .authenticate import Authenticate
from .receive import ReceiveData
from .entries import Entry, Entries
from .establish_connection import EstablishConnection
from .type_int import TypeInt
from .message import MessageReceived, MessageSend, MC_AUTO

logger = logging.getLogger(__name__)


class QueuedMessage:
    """Wrapper for a message to be sent along with its optional ACK callback."""

    def __init__(self, message: MessageSend, async_ack_callback=None):
        self.message = message
        self.async_ack_callback = async_ack_callback


KEEPALIVE_INTERVAL = 270.0
MINIMUM_MC = 20000  # minimum message count number
MAXIMUM_MC = 32000  # maximum message count number
HEADERS = {}


class Connection:
    def __init__(
        self,
        *,
        strict_mode: bool = False,
        message_log_filename: str | None = None,
        dump_entities: bool = False,
        traverse_screens_on_init: bool = False,
    ) -> None:
        self.ws = None
        self.strict = strict_mode
        self.message_log_filename = message_log_filename
        self.dump_entities = dump_entities
        self.traverse_screens_on_init = traverse_screens_on_init
        self.device_id = None
        self.device_version = None
        self.connection_id = None
        self.device_signature = None
        self.token = None
        self.device_home_id = None
        self.device_home_name = None
        self.sc_id = None
        self.async_notify_update_callbacks = []
        self.event_initial_setup_complete = asyncio.Event()
        self.event_initial_traverse_screens_complete = asyncio.Event()
        self.entries = Entries(self)
        self.data = ReceiveData(self)
        self.establish_connection_obj = EstablishConnection(self)
        self.authentication_obj = Authenticate(self)
        self.message_queue = asyncio.Queue()
        self.message_counter = MINIMUM_MC
        self.cipher = None
        self.pending_ack_callbacks = {}  # ref -> callback mapping

        # Create TraverseScreens object if screen traversal is enabled
        self.traverse_screens_obj = None
        if self.traverse_screens_on_init:
            from .traverse_screens import TraverseScreens

            self.traverse_screens_obj = TraverseScreens(self)
            # we need to add it here, before connection is being initialized
            # as it has to intercept messages that send over the initial screens
            self.add_async_notify_update_callback(
                self.traverse_screens_obj.entry_update_callback
            )

    def _log_raw_message(self, direction: str, payload: str) -> None:
        """Append raw message text to the log file, if configured."""
        if not self.message_log_filename:
            return
        try:
            # Parse JSON and pretty-print it
            try:
                parsed = json.loads(payload)
                pretty_payload = json.dumps(parsed, indent=2)
            except (json.JSONDecodeError, ValueError):
                # If not valid JSON, use as-is
                pretty_payload = payload

            with open(self.message_log_filename, "a", encoding="utf-8") as f:
                f.write(f"{direction}\n{pretty_payload}\n\n")
        except Exception:  # pragma: no cover - logging should not break flow
            logger.warning("Failed to write raw message log", exc_info=True)

    def add_async_notify_update_callback(self, callback) -> None:
        if callback not in self.async_notify_update_callbacks:
            self.async_notify_update_callbacks.append(callback)

    def remove_async_notify_update_callback(self, callback) -> None:
        if callback in self.async_notify_update_callbacks:
            self.async_notify_update_callbacks.remove(callback)

    async def async_notify_update(self, key: tuple, entry: Entry, new: bool) -> None:
        for callback in self.async_notify_update_callbacks:
            await callback(key, entry, new)

    async def initial_connection_setup_complete(self) -> None:
        """Called when initial connection setup is complete (after SET_ALERTS)."""
        if not self.event_initial_setup_complete.is_set():
            self.event_initial_setup_complete.set()
            # If traverse_screens is enabled, start traversal in background
            # Note: traverse_screens will wait for event_initial_setup_complete before proceeding anyway
            if self.traverse_screens_obj:

                async def run_traverse():
                    try:
                        await self.traverse_screens_obj.traverse_screens()
                    except Exception as e:
                        logger.error(
                            "Error during traverse_screens: %s", e, exc_info=True
                        )
                    finally:
                        self.remove_async_notify_update_callback(
                            self.traverse_screens_obj.entry_update_callback
                        )
                        self.event_initial_traverse_screens_complete.set()

                asyncio.create_task(run_traverse())

    async def enqueue_authentication_flow(
        self,
        username,
        password,
        auto_start_session=True,
        async_auth_result_callback=None,
    ):
        return await self.authentication_obj.enqueue_authentication_flow(
            username,
            password,
            auto_start_session,
            async_auth_result_callback,
        )

    async def receive_raw_message(self):
        buffer = b""
        while b"\x04" not in buffer:
            try:
                chunk = await self.ws.recv()  # type: ignore
            except (
                websockets.exceptions.ConnectionClosed,
                websockets.exceptions.ConnectionClosedOK,
            ) as e:
                logger.warning("Connection closed during recv: %s", e)
                raise

            if isinstance(chunk, str):
                chunk = chunk.encode("utf-8")
            if not chunk:
                raise ConnectionError("Connection closed before message complete")
            buffer += chunk

        if buffer.endswith(b"\x04"):
            msg_bytes = buffer[:-1]
        else:
            raise Exception(
                "Protocol error: message does not end with terminator", buffer
            )
        return msg_bytes

    async def receive_message(self):
        msg_encoded = await self.receive_raw_message()

        encrypted_data = base64.b64decode(msg_encoded)

        if self.cipher is None:
            raise RuntimeError("Cipher not initialized; secure connection missing")

        decryptor = self.cipher.decryptor()
        decrypted_padded = decryptor.update(encrypted_data) + decryptor.finalize()
        decrypted_msg = decrypted_padded.rstrip(b"\x00").decode("utf-8")

        self._log_raw_message("recv", decrypted_msg)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Received decrypted JSON: {decrypted_msg}")
        data = json.loads(decrypted_msg)

        if self.strict:
            allowed_keys = {"type_int", "mc", "payload", "ref"}
            extra_keys = set(data.keys()) - allowed_keys
            if extra_keys:
                logger.error(
                    "Strict mode: unexpected fields %s in message: %s", extra_keys, data
                )
                raise ValueError(f"Message contains unexpected fields: {extra_keys}")

        return MessageReceived(
            type_int=data.get("type_int"),
            mc=data.get("mc"),
            payload=data.get("payload"),
            ref=data.get("ref"),
        )

    async def messages_process(self):
        last_keepalive = asyncio.get_event_loop().time()
        socket_task: asyncio.Task | None = None
        queue_task: asyncio.Task | None = None

        try:
            while True:
                current_time = asyncio.get_event_loop().time()
                if current_time - last_keepalive >= KEEPALIVE_INTERVAL:
                    await self.send_keepalive()
                    last_keepalive = current_time

                time_until_keepalive = KEEPALIVE_INTERVAL - (
                    current_time - last_keepalive
                )
                timeout = max(0.1, time_until_keepalive)

                socket_task = asyncio.create_task(self.receive_message())
                queue_task = asyncio.create_task(self.message_queue.get())

                done, pending = await asyncio.wait(
                    [socket_task, queue_task],
                    timeout=timeout,
                    return_when=asyncio.FIRST_COMPLETED,
                )

                for task in pending:
                    task.cancel()
                    with suppress(asyncio.CancelledError):
                        await task

                queued_msg = None
                msg = None
                exception_to_raise = None

                for task in done:
                    try:
                        if task == queue_task:
                            queued_msg = task.result()
                        elif task == socket_task:
                            msg = task.result()
                    except asyncio.TimeoutError:
                        pass
                    except (
                        websockets.exceptions.ConnectionClosed,
                        websockets.exceptions.ConnectionClosedOK,
                    ) as e:
                        logger.warning("Connection closed detected in messages_process")
                        exception_to_raise = e
                    except Exception as e:  # pragma: no cover - defensive catch
                        logger.warning("Error receiving message: %s", e, exc_info=True)
                        exception_to_raise = e

                if exception_to_raise:
                    raise exception_to_raise

                if queued_msg is not None:
                    # Extract MessageSend and register callback if present
                    msg_to_send = queued_msg.message
                    # Auto-fill MC_AUTO with actual message counter
                    if msg_to_send.mc == MC_AUTO:
                        msg_to_send.mc = self.new_message_reference()
                    if queued_msg.async_ack_callback and msg_to_send.mc:
                        self.pending_ack_callbacks[msg_to_send.mc] = (
                            queued_msg.async_ack_callback
                        )
                    await self.send_message(msg_to_send)

                if msg is not None:
                    await self.async_dispatch_message(msg)

        except Exception:
            for task in (socket_task, queue_task):
                if task and not task.done():
                    task.cancel()
                    with suppress(asyncio.CancelledError):
                        await task
            raise

    async def send_ack_message(self, message_received):
        await self.send_message(
            MessageSend(
                type_int=TypeInt.ACK,
                ref=message_received.mc,
            )
        )
        logger.debug(f"Sent ACK for mc: {message_received.mc}")

    async def async_dispatch_message(self, msg: MessageReceived) -> None:
        if msg.type_int == TypeInt.ACK:
            await self.recv_ACK(msg)
        elif msg.type_int == TypeInt.NACK:
            await self.recv_NACK(msg)
        elif msg.type_int == TypeInt.HEARTBEAT:
            # Heartbeat messages are no-op, device is just keeping connection alive
            logger.debug("Received HEARTBEAT")
        elif msg.type_int == TypeInt.WAIT4ACK:
            # Device sends WAIT4ACK if it cannot answer the request within 3 seconds
            logger.debug(
                "Received WAIT4ACK, device is processing the request for more than 3 seconds"
            )
        elif msg.type_int == TypeInt.SET_HOME_DATA:
            await self.data.recv_SET_HOME_DATA(msg)
        elif msg.type_int == TypeInt.APP_INFO:
            await self.data.recv_APP_INFO(msg)
        elif msg.type_int == TypeInt.AUTH_LOGIN_DENIED:
            await self.authentication_obj.recv_AUTH_LOGIN_DENIED(msg)
        elif msg.type_int == TypeInt.AUTH_LOGIN_SUCCESS:
            await self.authentication_obj.recv_AUTH_LOGIN_SUCCESS(msg)
        elif msg.type_int == TypeInt.AUTH_APPLY_TOKEN_RESPONSE:
            await self.authentication_obj.recv_AUTH_APPLY_TOKEN_RESPONSE(msg)
        elif msg.type_int == TypeInt.ADD_SCREEN:
            await self.data.recv_ADD_SCREEN(msg)
        elif msg.type_int == TypeInt.ADD_DBENTRIES:
            await self.data.recv_ADD_DBENTRIES(msg)
        elif msg.type_int == TypeInt.ADD_ITEMS:
            await self.data.recv_ADD_ITEMS(msg)
        elif msg.type_int == TypeInt.SET_ALERTS:
            await self.data.recv_SET_ALERTS(msg)
            # there is no formal "initialization complete" message, but when alerts come, everything else seem to be already setup
            await self.initial_connection_setup_complete()
        elif msg.type_int == TypeInt.UPDATE_ITEMS:
            await self.data.recv_UPDATE_ITEMS(msg)
        else:
            logger.warning(f"Unhandled message: {msg}")

    async def recv_ACK(self, msg: MessageReceived) -> None:
        ref = msg.ref

        # Call callback if registered
        callback = self.pending_ack_callbacks.pop(ref, None)
        if callback is not None:
            await callback(True)

    async def recv_NACK(self, msg: MessageReceived) -> None:
        ref = msg.ref
        logger.info("Received NACK for ref: %s", ref)

        # Call callback if registered
        callback = self.pending_ack_callbacks.pop(ref, None)
        if callback is not None:
            await callback(False)

    async def send_message(self, msg: MessageSend) -> None:
        # Auto-fill MC_AUTO with actual message counter
        if msg.mc == MC_AUTO:
            msg.mc = self.new_message_reference()
        
        data = {
            "type_int": msg.type_int,
        }
        if msg.mc:
            data["mc"] = msg.mc
        if msg.payload:
            data["payload"] = msg.payload
        if msg.ref:
            data["ref"] = msg.ref

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Sending JSON: {data}")

        json_text = json.dumps(data)
        self._log_raw_message("send", json_text)

        msg_bytes = json_text.encode("utf-8")
        await self.encrypt_and_send_raw_message(msg_bytes)

    def new_message_reference(self) -> int:
        self.message_counter += 1
        if self.message_counter >= MAXIMUM_MC:
            self.message_counter = MINIMUM_MC
        return self.message_counter

    async def send_keepalive(self):
        msg = MessageSend(
            type_int=TypeInt.ACTION_SELECTED,
            payload={"ScreenID": 100},
            mc=self.new_message_reference(),
        )
        await self.send_message(msg)
        logger.debug("Sent ACTION_SELECTED with ScreenID=100")

    async def enqueue_message(self, msg: MessageSend, async_ack_callback=None) -> None:
        queued = QueuedMessage(msg, async_ack_callback)
        await self.message_queue.put(queued)

    async def enqueue_message_get_ack(self, msg: MessageSend) -> bool:
        """Enqueue a message and wait for ACK/NACK. Returns True for ACK, False for NACK."""
        ack_event = asyncio.Event()
        ack_result = False

        async def ack_callback(is_ack: bool):
            nonlocal ack_result
            ack_result = is_ack
            ack_event.set()

        queued = QueuedMessage(msg, ack_callback)
        await self.message_queue.put(queued)

        await ack_event.wait()
        return ack_result

    async def encrypt_and_send_raw_message(self, msg_bytes):
        pad_len = 16 - (len(msg_bytes) % 16)
        if pad_len < 16:
            msg_bytes += b"\x00" * pad_len

        if self.cipher is None:
            raise RuntimeError("Cipher not initialized; secure connection missing")

        encryptor = self.cipher.encryptor()
        encrypted_data = encryptor.update(msg_bytes) + encryptor.finalize()

        await self.ws.send(base64.b64encode(encrypted_data) + b"\x04")

    async def establish_secure_connection(self):
        return await self.establish_connection_obj.establish_secure_connection()

    async def async_websocket_connect(self, ws_url, proxy_url=None):
        ws_connect_kwargs = {
            "proxy": True,
        }

        if proxy_url:
            logger.info("Using proxy: %s", proxy_url)
            ws_connect_kwargs["proxy"] = proxy_url

        self.ws = await websockets.connect(ws_url, **ws_connect_kwargs)

        logger.info("Connected to WS server")

    async def async_close(self):
        if self.ws is not None:
            with suppress(Exception):
                await self.ws.close()
            self.ws = None



import asyncio
import logging

from .connection import Connection
from .entries import Entry


class TraverseScreens:
    """Traverses all screens. Act as a very disciplined human going through all the menues.
    The issue is that some screens we get to know about ahead of time (sent by the device at
    initialization of the connection), but for others we actually need to enter them to get
    their descriptions

    The logics is that if it's not a back button, and it's not an action that does something,
    we should try clicking and seeing what happens.

    Hopefully we have the right conditions for those:
    detail: 0 => back button
    detail: 1 => action that changes something -- but we don't know what as this is business logic
    on the device's side.
    All other "buttons" presumably lead to screens.
    """

    def __init__(self, conn_obj: Connection) -> None:
        self.conn_obj = conn_obj
        self.explore_queue: asyncio.Queue = asyncio.Queue()
        self.requested_screens: set = set()

    async def entry_update_callback(self, key: tuple, entry: Entry, new: bool) -> None:
        """Process updates as they arrive from the server."""
        # vid 317 is system time
        # vid 318 is "a problem"
        if key == (0, 0, None) or key == (317, 1, None) or key == (318, 1, None):
            return
        update_type = "NEW" if new else "UPDATE"
        # print(f"[{update_type}] {key}: {entry}")

        # Log entities with detail = 0 or detail = 1 to a separate file
        detail = key[1]
        _vid_obj = entry.get("_vid_obj", {})
        # if new and detail in (0, 1) and _vid_obj.get("type") == 1:
        #    with open("detail_0_1_entities.txt", "a", encoding="utf-8") as f:
        #        f.write(f"{update_type}: {key}: {entry}\n")

        # Check if this is a type 1 entry and add to explore queue (but exclude detail=1)
        if new and detail not in (0, 1):
            if _vid_obj.get("type") == 1:
                await self.explore_queue.put((key, entry))
                logging.debug(f"Added entry to explore queue: {key}")

    async def traverse_screens(self) -> None:
        """Traverse screens by requesting details for type 1 entries from the explore queue."""
        logging.info("Waiting for connection initialization to complete...")
        await self.conn_obj.event_initial_setup_complete.wait()
        logging.info(
            "Initialization complete. Traversing all the screens to do discovery."
        )

        # Continuously pull items from the explore queue
        while True:
            try:
                key, entry = await asyncio.wait_for(
                    self.explore_queue.get(), timeout=3.0
                )
            except asyncio.TimeoutError:
                logging.info("traverse_screens(): queue idle; exiting traversal")
                break

            try:
                vid, detail, obj_id = key  # unpack tuple
                originating_screen_key = entry.get("_screen_key")
                if not originating_screen_key:
                    logging.error(
                        f"No originating_screen_key found for entry {key}, skipping"
                    )
                    continue

                # Create ACTUAL_SCREEN message to navigate to the originating screen
                screen_msg = entry.message_activate_entering_screen()
                actual_screen_ack = await self.conn_obj.enqueue_message_get_ack(
                    screen_msg
                )
                # logging.info(
                #    f"Received ACK for ACTUAL_SCREEN {originating_screen_key},  succes: {actual_screen_ack}"
                # )
                # Create and enqueue the ACTION_SELECTED message
                msg = entry.message_action_selected()
                # POTENTIAL BUG : these actions can trigger additional entry_update_callback calls as the device responds.
                # ASSESMENT: IT'S OK as this is run immediately after connection initialization, before

                logging.debug(f"Enqueued ACTION_SELECTED for {key}: {msg.payload}")
                action_selected_ack = await self.conn_obj.enqueue_message_get_ack(msg)
                logging.debug(
                    f"Received ACK for ACTION_SELECTED {key}, succes: {action_selected_ack}"
                )
            except asyncio.CancelledError:
                logging.info("Traverse_screens() processing CanceledError exception.")
                raise

        logging.info("traverse_screens() complete")



import asyncio
import base64
import json
import unittest
from unittest.mock import patch

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from harreither_brain_client.establish_connection import EstablishConnection
from harreither_brain_client.type_int import TypeInt


class FakeWebSocket:
    def __init__(self) -> None:
        self.sent: list[str] = []

    async def send(self, data: str) -> None:
        self.sent.append(data)


class FakeConnection:
    def __init__(self, recv_bytes: list[bytes]) -> None:
        self.ws = FakeWebSocket()
        self._recv_iter = iter(recv_bytes)
        self.device_id = None
        self.device_version = None
        self.connection_id = None
        self.public_key = None
        self.device_signature = None
        self.session_key = None
        self.session_iv = None
        self.cipher = None
        self.sc_id = None

    async def receive_raw_message(self) -> bytes:
        try:
            return next(self._recv_iter)
        except StopIteration:
            raise AssertionError("No more fake messages queued")


class EstablishConnectionTests(unittest.IsolatedAsyncioTestCase):
    async def test_establish_secure_connection_happy_path(self):
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_pem = (
            private_key.public_key()
            .public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            .decode("utf-8")
        )

        session_key = b"K" * 32
        session_iv = b"I" * 16

        sc_established_payload = json.dumps(
            {"type_int": TypeInt.SC_ESTABLISHED, "payload": {"sc_id": "secure-1"}}
        ).encode("utf-8")
        pad_len = (16 - (len(sc_established_payload) % 16)) % 16
        padded = sc_established_payload + b"\x00" * pad_len
        encryptor = Cipher(
            algorithms.AES(session_key),
            modes.CBC(session_iv),
            backend=default_backend(),
        ).encryptor()
        resp17_ciphertext = encryptor.update(padded) + encryptor.finalize()
        resp17_base64 = base64.b64encode(resp17_ciphertext)

        first = json.dumps(
            {
                "type_int": TypeInt.CONNECTION_START,
                "payload": {
                    "device_id": "dev-1",
                    "device_version": "1.0",
                    "connection_id": "conn-1",
                },
            }
        ).encode("utf-8")
        resp_11 = json.dumps({"type_int": TypeInt.CONNECTION_ESTABLISHED}).encode(
            "utf-8"
        )
        resp_14 = json.dumps(
            {
                "type_int": TypeInt.SC_PUBKEY,
                "payload": {"public_key": public_pem, "device_signature": "sig"},
            }
        ).encode("utf-8")

        fake_conn = FakeConnection([first, resp_11, resp_14, resp17_base64])
        handler = EstablishConnection(fake_conn)

        with patch(
            "harreither_brain_client.establish_connection.token_bytes",
            side_effect=[session_key, session_iv],
        ):
            await handler.establish_secure_connection()

        self.assertEqual(fake_conn.device_id, "dev-1")
        self.assertEqual(fake_conn.device_version, "1.0")
        self.assertEqual(fake_conn.connection_id, "conn-1")

        secret_msg = json.loads(fake_conn.ws.sent[2][:-1])
        encrypted_secret_b64 = secret_msg["payload"]["secret"]
        decrypted_secret = private_key.decrypt(
            base64.b64decode(encrypted_secret_b64), padding.PKCS1v15()
        ).decode("utf-8")
        self.assertEqual(decrypted_secret, f"{session_key.hex()}:::{session_iv.hex()}")

        self.assertIsNotNone(fake_conn.cipher)
        self.assertEqual(fake_conn.session_key, session_key)
        self.assertEqual(fake_conn.session_iv, session_iv)
        self.assertEqual(fake_conn.sc_id, "secure-1")

        sent_msgs = [json.loads(msg[:-1]) for msg in fake_conn.ws.sent]
        self.assertEqual(sent_msgs[0]["type_int"], TypeInt.CONNECTION_CONFIRM)
        self.assertEqual(sent_msgs[1]["type_int"], TypeInt.SC_INIT)
        self.assertEqual(sent_msgs[2]["type_int"], TypeInt.SC_SECRET)

    async def test_establish_secure_connection_declined_raises(self):
        first = json.dumps({"type_int": TypeInt.CONNECTION_START, "payload": {}}).encode(
            "utf-8"
        )
        declined = json.dumps(
            {"type_int": TypeInt.CONNECTION_DECLINED, "payload": {"reason": "nope"}}
        ).encode("utf-8")
        fake_conn = FakeConnection([first, declined])
        handler = EstablishConnection(fake_conn)

        with self.assertRaises(ConnectionError):
            await handler.establish_secure_connection()

        sent_msgs = [json.loads(msg[:-1]) for msg in fake_conn.ws.sent]
        self.assertEqual(len(sent_msgs), 1)
        self.assertEqual(sent_msgs[0]["type_int"], TypeInt.CONNECTION_CONFIRM)

    async def test_establish_secure_connection_wrong_first_message(self):
        first = json.dumps({"type_int": TypeInt.HEARTBEAT}).encode("utf-8")
        fake_conn = FakeConnection([first])
        handler = EstablishConnection(fake_conn)

        with self.assertRaises(ValueError):
            await handler.establish_secure_connection()

        self.assertEqual(fake_conn.ws.sent, [])

    async def test_establish_secure_connection_missing_pubkey(self):
        first = json.dumps({"type_int": TypeInt.CONNECTION_START, "payload": {}}).encode(
            "utf-8"
        )
        resp_11 = json.dumps({"type_int": TypeInt.CONNECTION_ESTABLISHED}).encode(
            "utf-8"
        )
        resp_14 = json.dumps(
            {
                "type_int": TypeInt.SC_PUBKEY,
                "payload": {"public_key": None, "device_signature": None},
            }
        ).encode("utf-8")
        fake_conn = FakeConnection([first, resp_11, resp_14])
        handler = EstablishConnection(fake_conn)

        with self.assertRaises(ValueError):
            await handler.establish_secure_connection()

        sent_msgs = [json.loads(msg[:-1]) for msg in fake_conn.ws.sent]
        self.assertEqual(len(sent_msgs), 2)
        self.assertEqual(sent_msgs[0]["type_int"], TypeInt.CONNECTION_CONFIRM)
        self.assertEqual(sent_msgs[1]["type_int"], TypeInt.SC_INIT)


if __name__ == "__main__":
    unittest.main()



import base64
import json
import logging
from secrets import token_bytes

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .type_int import TypeInt

logger = logging.getLogger(__name__)


class EstablishConnection:  # noqa: D101
    def __init__(self, connection) -> None:
        self.connection = connection
        self.public_key: rsa.RSAPublicKey | None = None
        self.session_key: bytes | None = None
        self.session_iv: bytes | None = None

    async def recv_unencrypted(self, label):
        msg_bytes = await self.connection.receive_raw_message()
        return json.loads(msg_bytes.decode("utf-8"))

    async def send_json_message(self, type_int, payload=None, mc=-1):
        data = {
            "type_int": type_int,
            "mc": mc,
        }
        if payload:
            data["payload"] = payload

        await self.connection.ws.send(json.dumps(data) + "\x04")
        logger.debug(f"Sent type_int: {type_int} ({TypeInt.get_name(type_int)})")

    async def establish_secure_connection(self):
        first = await self.recv_unencrypted("FIRST")

        if first.get("type_int") != TypeInt.CONNECTION_START:
            raise ValueError(
                f"Expected CONNECTION_START ({TypeInt.CONNECTION_START.value}), got {first.get('type_int')}"
            )

        payload = first.get("payload", {})
        self.connection.device_id = payload.get("device_id")
        self.connection.device_version = payload.get("device_version")
        self.connection.connection_id = payload.get("connection_id")

        logger.info(
            f"Connected [10]. ID: {self.connection.connection_id}, Device: {self.connection.device_id}, Version: {self.connection.device_version}"
        )

        await self.send_json_message(
            type_int=TypeInt.CONNECTION_CONFIRM,
            payload={
                "client_type": "nextgen-app",
                "client_id": "abaf877411be6b59",
                "client_version": "1.2.2",
                "connection_id": self.connection.connection_id,
            },
        )

        resp_11 = await self.recv_unencrypted("RESP 11")
        resp_11_type = resp_11.get("type_int")

        if resp_11_type == TypeInt.CONNECTION_DECLINED:
            reason = resp_11.get("payload", {}).get("reason") or "unknown reason"
            raise ConnectionError(f"Connection declined by device: {reason}")

        if resp_11_type != TypeInt.CONNECTION_ESTABLISHED:
            raise ValueError(
                f"Expected CONNECTION_ESTABLISHED ({TypeInt.CONNECTION_ESTABLISHED.value}), got {resp_11_type}"
            )

        logger.info(f"Connection established [12]: {resp_11}")

        await self.send_json_message(type_int=TypeInt.SC_INIT)

        resp_14 = await self.recv_unencrypted("RESP 14")

        if resp_14.get("type_int") != TypeInt.SC_PUBKEY:
            raise ValueError(
                f"Expected SC_PUBKEY ({TypeInt.SC_PUBKEY.value}), got {resp_14.get('type_int')}"
            )

        payload = resp_14.get("payload", {})

        public_key = payload.get("public_key")
        device_signature = payload.get("device_signature")

        if not public_key or not device_signature:
            raise ValueError("Missing public_key or device_signature in payload")

        logger.debug(f"Public key [15]: {public_key}")
        logger.debug(f"Device signature [15]: {device_signature}")

        # in theory we would now check the device signature, but why...

        self.public_key = serialization.load_pem_public_key(
            public_key.encode("utf-8"), backend=default_backend()
        )

        self.connection.device_signature = device_signature

        self.session_key = token_bytes(32)
        self.session_iv = token_bytes(16)

        logger.debug(f"Generated Session Key: {self.session_key.hex()}")
        logger.debug(f"Generated Session IV: {self.session_iv.hex()}")

        # Only the cipher is needed by the connection for subsequent encrypted traffic
        self.connection.cipher = Cipher(
            algorithms.AES(self.session_key),
            modes.CBC(self.session_iv),
            backend=default_backend(),
        )
        secret_payload = (
            f"{self.session_key.hex()}:::{self.session_iv.hex()}"
        )

        encrypted_secret = self.public_key.encrypt(
            secret_payload.encode("utf-8"), padding.PKCS1v15()
        )

        secret = base64.b64encode(encrypted_secret).decode("utf-8")
        await self.send_json_message(
            type_int=TypeInt.SC_SECRET,
            payload={"secret": secret},
        )
        msg_bytes_17 = await self.connection.receive_raw_message()
        resp_17_raw = base64.b64decode(msg_bytes_17)
        decryptor = self.connection.cipher.decryptor()
        decrypted_17_padded = decryptor.update(resp_17_raw) + decryptor.finalize()
        decrypted_17_bytes = decrypted_17_padded.rstrip(b"\x00")

        try:
            resp_17 = json.loads(decrypted_17_bytes.decode("utf-8"))
        except json.JSONDecodeError:
            raise ValueError(
                f"Failed to decode SC_ESTABLISHED message as JSON: {decrypted_17_bytes}"
            )

        if resp_17.get("type_int") != TypeInt.SC_ESTABLISHED:
            raise ValueError(
                f"Expected SC_ESTABLISHED ({TypeInt.SC_ESTABLISHED.value}), got {resp_17.get('type_int')}"
            )

        payload = resp_17.get("payload", {})
        self.connection.sc_id = payload.get("sc_id")

        logger.info(f"Secure Connection established. SC_ID: {self.connection.sc_id}")



import json
import logging
from typing import Any

logger = logging.getLogger(__name__)


class ReceiveData:
    def __init__(self, connection):
        self.connection = connection

    def validate_keys(self, data, allowed_keys, context):
        if self.connection.strict:
            extra_keys = set(data.keys()) - allowed_keys
            if extra_keys:
                extra_fields = {k: data[k] for k in extra_keys}
                logger.error(
                    "Strict mode: %s contains unexpected fields %s; full payload: %s",
                    context,
                    extra_keys,
                    data,
                )
                raise ValueError(
                    f"{context} contains unexpected fields: {extra_fields}"
                )

    async def recv_SET_HOME_DATA(self, message):
        payload = message.payload

        allowed_keys = {"id", "name", "isvalid", "info"}
        self.validate_keys(payload, allowed_keys, "SET_HOME_DATA payload")

        self.connection.device_home_id = payload.get("id")
        self.connection.device_home_name = payload.get("name")
        logger.info(
            f"Received SET_HOME_DATA [301]. Payload size: {len(str(message.payload))}"
        )

        await self.connection.send_ack_message(message)

    async def recv_APP_INFO(self, message):
        payload = message.payload
        self.validate_keys(payload, {"info"}, "APP_INFO payload")
        logger.info(
            f"Received APP_INFO [295]. Payload size: {len(str(message.payload))}"
        )
        await self.connection.send_ack_message(message)

    async def recv_ADD_SCREEN(self, message):
        logger.info(
            f"Received ADD_SCREEN [296]. Payload size: {len(str(message.payload))}"
        )

        payload = message.payload
        self.validate_keys(payload, {"screen"}, "ADD_SCREEN payload")

        if "screen" not in payload:
            raise ValueError("ADD_SCREEN message missing 'screen' dictionary")
        screen_data = payload["screen"]

        allowed_keys = {
            "screenID",
            "title",
            "statuspage",
            "itemCount",
            "objID",
            "iconID",
        }
        self.validate_keys(screen_data, allowed_keys, "Screen dictionary")

        screen_id = screen_data.get("screenID")
        obj_id = screen_data.get("objID")
        if screen_id is not None:
            self.connection.entries.screens[(screen_id, obj_id)] = screen_data

        self.connection.entries.save_screens()

        await self.connection.send_ack_message(message)

    async def recv_ADD_DBENTRIES(self, message):
        logger.info(
            f"Received ADD_DBENTRIES [297]. Payload size: {len(str(message.payload))}"
        )

        payload = message.payload
        self.validate_keys(payload, {"DBentries"}, "ADD_DBENTRIES payload")
        db_entries = payload.get("DBentries", [])

        allowed_keys = {
            "VID",
            "type",
            "text",
            "min",
            "max",
            "step",
            "elements",
            "unit",
            "pwd",
        }

        for entry in db_entries:
            self.validate_keys(entry, allowed_keys, "DB entry")

            vid = entry.get("VID")
            if vid is not None:
                self.connection.entries.dbentries[vid] = entry

        if self.connection.dump_entities:
            with open("dbentries.json", "w", encoding="utf-8") as f:
                json.dump(self.connection.entries.dbentries, f, indent=4, ensure_ascii=False)

        await self.connection.send_ack_message(message)

    async def recv_ADD_ITEMS(self, message):
        logger.info(
            f"Received ADD_ITEMS [299]. Payload size: {len(str(message.payload))}"
        )
        payload = message.payload
        self.validate_keys(
            payload,
            {
                "screenID",
                "pos",
                "items",
                "end",
                "objID",
            },
            "ADD_ITEMS payload",
        )
        screen_id = payload.get("screenID")
        obj_id = payload.get("objID")

        screen_key = (screen_id, obj_id)

        if screen_key not in self.connection.entries.screens:
            logger.warning(
                f"Received ADD_ITEMS for unknown screenID: {screen_id}. Initializing screen entry."
            )
            self.connection.entries.screens[screen_key] = {"screenID": screen_id, "objID": obj_id}

        if "items" not in self.connection.entries.screens[screen_key]:
            self.connection.entries.screens[screen_key]["items"] = []

        def fix_item(it):
            allowed_keys = {
                "VID",
                "detail",
                "name",
                "edit",
                "value",
                "validity",
                "valstr",
                "objID",
                "citems",
                "highlight",
            }
            self.validate_keys(it, allowed_keys, "Item")

            vid = it.get("VID")
            if vid is not None:
                _vid_obj = self.connection.entries.dbentries.get(vid)
                if _vid_obj is not None:
                    it["_vid_obj"] = _vid_obj

            # Add screen key reference
            it["_screen_key"] = screen_key

            if "citems" in it and isinstance(it["citems"], list):
                it["citems"] = [fix_item(cit) for cit in it["citems"]]
            return it

        new_items = [fix_item(it) for it in payload.get("items", [])]
        await self.add_entries(new_items)
        self.connection.entries.screens[screen_key]["items"].extend(new_items)

        self.connection.entries.save_screens()

        logger.info(
            f"Updated screen {screen_id} with {len(new_items)} items. Total: {len(self.connection.entries.screens[screen_key].get('items', []))}"
        )
        await self.connection.send_ack_message(message)

    async def recv_SET_ALERTS(self, message):
        logger.debug(
            f"Received SET_ALERTS [302]. Payload size: {len(str(message.payload))}"
        )
        payload = message.payload
        self.validate_keys(
            payload, {"restart", "alerts", "end"}, "SET_ALERTS payload"
        )

        if payload.get("restart"):
            self.connection.entries.alerts = []

        new_alerts = payload.get("alerts", [])
        allowed_keys = {"text", "type", "icon"}

        for alert in new_alerts:
            self.validate_keys(alert, allowed_keys, "Alert")

            text = alert.get("text")
            icon = alert.get("icon")
            logger.info(f"Alert Received: {text} (Icon: {icon})")

            self.connection.entries.alerts.append(alert)

        await self.connection.send_ack_message(message)

    async def recv_UPDATE_ITEMS(self, message):
        logger.debug(
            f"Received UPDATE_ITEMS [300]. Payload size: {len(str(message.payload))}"
        )
        payload = message.payload
        self.validate_keys(payload, {"items", "end"}, "UPDATE_ITEMS payload")
        update_items = payload.get("items", [])

        for u_item in update_items:
            key = self.connection.entries.make_key_from_object(u_item)
            await self.connection.entries.update_entry(key, u_item)

        self.connection.entries.save_entries()
        self.connection.entries.save_screens()

        await self.connection.send_ack_message(message)

    async def add_entries(self, items, objID=None, _top_level=True):
        for item in items:
            vid = item.get("VID")
            if vid is not None:
                _vid_obj = self.connection.entries.dbentries.get(vid)
                if _vid_obj is not None:
                    item["_vid_obj"] = _vid_obj

            if "detail" in item:
                if objID:
                    item["objID"] = objID
                if "citems" in item and isinstance(item["citems"], list):
                    if item["citems"][0].get("objID", None):
                        item["objID"] = item["citems"][0]["objID"]

                key = self.connection.entries.make_key_from_object(item)
                await self.connection.entries.create_entry(key, item)

            if "citems" in item and isinstance(item["citems"], list):
                await self.add_entries(item["citems"], objID=None, _top_level=False)

        if _top_level:
            self.connection.entries.save_entries()



import copy
import json
import logging
from typing import Any, Tuple

from .message import MessageSend, MC_AUTO
from .type_int import TypeInt

logger = logging.getLogger(__name__)


class Entry(dict):
    """A dict-like entry with built-in normalization."""

    def normalize(self) -> Any:
        """Deep-copy self while removing 'name', 'highlight', and '_screen_key' fields for comparisons."""
        return self._normalize_value(dict(self))

    @staticmethod
    def _normalize_value(value: Any) -> Any:
        """Helper method to normalize a dict-representation of an entry recursively."""
        if isinstance(value, dict):
            # Remove name, highlight, and _screen_key fields if present
            value = copy.deepcopy(value)
            value.pop("name", None)  # we should not be doing that
            value.pop("highlight", None)
            value.pop("_screen_key", None)
            # Recursively normalize nested values
            for k, v in value.items():
                value[k] = Entry._normalize_value(v)
            return value
        if isinstance(value, list):
            return [Entry._normalize_value(v) for v in value]
        return value

    def message_edit_value(self, value: int) -> MessageSend:
        """Create a SELECT message to change the device value.

        Args:
            value: The index of the selected option (0-based).

        Returns:
            A MessageSend object for the ACTION_EDITED_VALUE message.
        """
        vid = self["VID"]
        detail = self["detail"]
        objid = self.get("objID")

        payload: dict[str, Any] = {
            "VID": vid,
            "detail": detail,
            "value": value,
            "validity": 0,
        }
        if objid is not None:
            payload["objID"] = objid

        return MessageSend(
            type_int=TypeInt.ACTION_EDITED_VALUE,
            mc=MC_AUTO,
            payload=payload,
        )

    def message_activate_entering_screen(self) -> MessageSend:
        """Create a message to activate the screen associated with this entry.

        Returns:
            A MessageSend object for the ACTUAL_SCREEN message.
        """
        originating_screen_key = self.get("_screen_key")
        if not originating_screen_key:
            raise ValueError(f"Entry {self} has no _screen_key")

        screen_id, screen_obj_id = originating_screen_key
        screen_payload: dict[str, Any] = {"screenID": screen_id}
        if screen_obj_id is not None:
            screen_payload["objID"] = screen_obj_id

        return MessageSend(
            type_int=TypeInt.ACTUAL_SCREEN,
            mc=MC_AUTO,
            payload=screen_payload,
        )

    def message_action_selected(self) -> MessageSend:
        """Create a message to trigger ACTION_SELECTED for this entry.
            You need to activate the screen before that for this to work properly!

        Returns:
            A MessageSend object for the ACTION_SELECTED message.
        """
        vid = self["VID"]
        detail = self["detail"]
        objid = self.get("objID")

        payload: dict[str, Any] = {"VID": vid, "detail": detail}
        if objid is not None:
            payload["objID"] = objid

        return MessageSend(
            type_int=TypeInt.ACTION_SELECTED,
            mc=MC_AUTO,
            payload=payload,
        )


class Entries:
    def __init__(self, connection):
        self._entries = {}
        self.connection = connection
        self.dbentries = {}
        self.screens = {}
        self.alerts = []

    def make_key_from_object(self, object) -> Tuple[Any, Any, Any]:
        return (object.get("VID"), object["detail"], object.get("objID", None))

    def get_entry(self, key: Tuple[Any, Any, Any]) -> Entry | None:
        return self._entries.get(key)

    async def create_entry(self, key, entry_data):
        # if key[1] == 0:
        #    return
        new = True

        # Create Entry object from input data
        entry = Entry(entry_data)

        if key in self._entries:
            _vid_obj = entry.get("_vid_obj", {})

            # Only enforce duplicate checks in strict mode
            if self.connection.strict and _vid_obj.get("type") not in (23, 4):
                existing_norm = self._entries[key].normalize()
                incoming_norm = entry.normalize()

                if existing_norm != incoming_norm:
                    raise ValueError(
                        f"create_entry() Entry {key} already exists, current data: {dict(self._entries[key])}, new data: {entry_data}"
                    )
                logger.debug(
                    "create_entry() Entry %s already exists with identical content (ignoring some fields), overwriting",
                    key,
                )
                new = False
        self._entries[key] = entry
        await self.connection.async_notify_update(key, entry, new)

    async def update_entry(self, key, updated_entry):
        if key == (0, 0, None):
            return
        if key == (318, 1, None):
            logger.info(f"update_entry(): Got a update of an entry that does not exist, likely 318 is an error message: {key}, entry: {updated_entry}")
            return
        entry = self.get_entry(key)
        if entry is None:
            raise ValueError(f"update_entry() Entry {key} does not exist")

        vid = key[0]
        vid_info = self.dbentries[vid]
        vid_text = vid_info["text"]
        vid_text = vid_info.get("text", f"VID:{vid}")
        entry_name = entry.get("name", "")

        for k, v in updated_entry.items():
            old_value = entry.get(k)
            if old_value != v:
                old_value_repr = repr(old_value)
                v_repr = repr(v)
                logger.debug(
                    f"Entry {vid_text} {entry_name} ({key}) field '{k}': {old_value_repr} -> {v_repr}"
                )
                entry[k] = v

        await self.connection.async_notify_update(key, entry, False)

    def save_entries(self):
        dumpable = {str(k): dict(v) for k, v in self._entries.items()}
        if self.connection.dump_entities:
            with open("entries.json", "w", encoding="utf-8") as f:
                json.dump(dumpable, f, indent=4, ensure_ascii=False)

    def save_screens(self):
        """Save screens dict with tuple keys to JSON, converting keys to strings and removing _screen_key."""
        if self.connection.dump_entities:
            # Helper to recursively remove _screen_key from items
            def remove_screen_key(obj):
                if isinstance(obj, dict):
                    result = {
                        k: remove_screen_key(v)
                        for k, v in obj.items()
                        if k != "_screen_key"
                    }
                    return result
                elif isinstance(obj, list):
                    return [remove_screen_key(item) for item in obj]
                return obj

            # Clean screens data
            cleaned_screens = remove_screen_key(self.screens)
            dumpable = {str(k): v for k, v in cleaned_screens.items()}
            with open("screens.json", "w", encoding="utf-8") as f:
                json.dump(dumpable, f, indent=4, ensure_ascii=False)



import argparse
import asyncio
import logging
import sys

from .connection import Connection
from .message import MessageSend
from .type_int import TypeInt


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Harreither Brain CLI Client")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose logging"
    )
    parser.add_argument(
        "-p", "--proxy", action="store_true", help="Use proxy on localhost:8080"
    )
    parser.add_argument(
        "--username",
        required=True,
        help="Username for authentication",
    )
    parser.add_argument(
        "--password",
        required=True,
        help="Password for authentication",
    )
    parser.add_argument(
        "--host",
        required=True,
        help="WebSocket server URL",
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="Fail on any unexpected fields from server",
    )
    parser.add_argument(
        "--logfile",
        help="If set, log raw messages to this file",
    )
    parser.add_argument(
        "--dumpentities",
        action="store_true",
        help="Dump dbentries, entries, and screens JSON files to disk",
    )
    parser.add_argument(
        "-a",
        action="store_true",
        help="Send single ACTION_SELECTED sequence variant A",
    )
    return parser


def configure_logging(verbose: bool) -> None:
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        stream=sys.stdout,
    )
    
    # Suppress debug logging for websockets even in verbose mode
    if verbose:
        logging.getLogger("websockets.client").setLevel(logging.INFO)


def get_entry_from_payload(conn_obj: Connection, payload: dict):
    """Get entry from payload by converting to key tuple."""
    key = (payload.get("VID"), payload.get("detail"), payload.get("objID"))
    return conn_obj.data.entries.get_entry(key)


async def send_single_action_a(conn_obj: Connection) -> None:
    """Send a sequence of actions and wait for ACK/NACK."""
    logging.info("Waiting for traversal to complete...")
    await conn_obj.event_initial_traverse_screens_complete.wait()
    logging.info("Traversal complete. Sending action sequence...")
    '''
    # 1. Activate screenID: 200
    screen_msg = MessageSend(
        type_int=TypeInt.ACTUAL_SCREEN,
        mc=conn_obj.new_message_reference(),
        payload={"screenID": 200},
    )
    logging.info(f"Sending ACTUAL_SCREEN with payload: {screen_msg.payload}")
    screen_ack = await conn_obj.enqueue_message_get_ack(screen_msg)
    logging.info(f"Received ACK for screen activation: {screen_ack}")
    await asyncio.sleep(1)
    
    # 2. Send ACTION_SELECTED message for entity {'VID': 30044, 'detail': 1012}
    action_msg1 = MessageSend(
        type_int=TypeInt.ACTION_SELECTED,
        mc=conn_obj.new_message_reference(),
        payload={"VID": 30044, "detail": 1012},
    )
    logging.info(f"Sending ACTION_SELECTED with payload: {action_msg1.payload}")
    action_ack1 = await conn_obj.enqueue_message_get_ack(action_msg1)
    logging.info(f"Received ACK for ACTION_SELECTED: {action_ack1}")
    await asyncio.sleep(1)
    
    # 3. Send ACTION_SELECTED message for entity {'VID': 3, 'detail': 1013, 'objID': 30018}
    action_msg2 = MessageSend(
        type_int=TypeInt.ACTION_SELECTED,
        mc=conn_obj.new_message_reference(),
        payload={"VID": 3, "detail": 1013, "objID": 30018},
    )
    logging.info(f"Sending ACTION_SELECTED with payload: {action_msg2.payload}")
    action_ack2 = await conn_obj.enqueue_message_get_ack(action_msg2)
    logging.info(f"Received ACK for ACTION_SELECTED: {action_ack2}")
    await asyncio.sleep(1)
    
    # 4. Activate entering screen for VID: 1161, detail: 30018
    entry = conn_obj.data.entries.get_entry((1161, 30018, None))
    if not entry:
        # Try to find with any objID
        for key, e in conn_obj.data.entries._entries.items():
            if key[0] == 1161 and key[1] == 30018:
                entry = e
                break
        if not entry:
            raise ValueError("Entry with VID=1161, detail=30018 not found after traversal")
    
    '''

    entry = conn_obj.data.entries.get_entry((1161, 30018, None))
    '''
    screen_msg = entry.message_activate_entering_screen(conn_obj)
    logging.info(f"Sending message_activate_entering_screen with payload: {screen_msg.payload}")
    screen_ack = await conn_obj.enqueue_message_get_ack(screen_msg)
    logging.info(f"Received ACK: {screen_ack}")
    '''

    msg = entry.message_edit_value(0, conn_obj)
    logging.info(f"Sending message_edit_value with payload: {msg.payload}")
    ack = await conn_obj.enqueue_message_get_ack(msg)
    logging.info(f"Received ACK: {ack}")
    await asyncio.sleep(1)
  


async def run_client(
    *,
    host: str,
    username: str,
    password: str,
    proxy_url: str | None,
    strict: bool,
    logfile: str | None,
    use_single_action_a: bool,
    dump_entities: bool,
) -> None:
    # Default mode (no single action) means we traverse screens
    traverse_screens_on_init = True
    conn_obj = Connection(
        strict_mode=strict,
        message_log_filename=logfile,
        dump_entities=dump_entities,
        traverse_screens_on_init=traverse_screens_on_init,
    )
    
    try:
        await conn_obj.async_websocket_connect(host, proxy_url=proxy_url)
        logging.info("Connected to WS server")

        await conn_obj.establish_secure_connection()
        await conn_obj.enqueue_authentication_flow(
            username=username,
            password=password,
            auto_start_session=True,
            async_auth_result_callback=None,
        )
        
        # Run messages_process and either send a single-action variant or request type 1 entries
        if use_single_action_a:
            await asyncio.gather(
                conn_obj.messages_process(),
                send_single_action_a(conn_obj),
            )
        else:
            await asyncio.gather(
                conn_obj.messages_process(),
            )

    except asyncio.CancelledError:
        logging.info("Connection task cancelled")
        raise
    except Exception as exc:  # pragma: no cover - top-level protection
        logging.error("Error during connection: %s", exc, exc_info=True)
    finally:
        await conn_obj.async_close()


def main(argv: list[str] | None = None) -> None:
    parser = build_parser()
    args = parser.parse_args(args=argv)

    configure_logging(args.verbose)

    proxy_url = "http://localhost:8080" if args.proxy else None

    asyncio.run(
        run_client(
            host=args.host,
            username=args.username,
            password=args.password,
            proxy_url=proxy_url,
            strict=args.strict,
            logfile=args.logfile,
            use_single_action_a=args.a,
            dump_entities=args.dumpentities,
        )
    )

if __name__ == "__main__":
    main()
